/**
 * Example shows how sign PDF file (GOST 34.10 algorithm) and verify signature and certificate.
 */
package tests;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CRL;
import java.security.cert.CRLException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import org.bouncycastle.tsp.TimeStampToken;
import com.itextpdf.text.DocumentException;
import com.itextpdf.text.pdf.AcroFields;
import com.itextpdf.text.pdf.PdfPKCS7;
import com.itextpdf.text.pdf.PdfReader;
import com.itextpdf.text.pdf.PdfSignatureAppearance;
import com.itextpdf.text.pdf.PdfStamper;

public class CryptoProPDF {
	
	// Path to PDF files
	final static String PREFIX_PDF_DIR = "/home/cross/Downloads/";
	// Path to certificate, CRL, CA, key store
	final static String PREFIX_KEY_DIR = "/dailybuilds/CSPbuild/rtests/cades/";
	
	// Alias of key in client store
	final static String ALIAS = "UnixCadesTestCerts-GostUsr";
	// Password to client store
	final static char[] PASSWORD = null;
	
	// Alias of CA certificate in trusted store
	final static String CA_ALIAS = "gost.cer";
	// Password to trusted store
	final static char[] CA_PASSWORD = "security".toCharArray();
	// CA file
	final static String CA_FILENAME =  "certstore.crs";
	
	// PDF file
	final static String PDF_FILENAME = "test.pdf";
	// CRL file
	final static String CRL_FILENAME = "root.crl";
		
	/**
	 * Loads key store.
	 * @param storeFile - key store file.
	 * @param password - password to store.
	 * @return opened key store.
	 * @throws KeyStoreException
	 * @throws NoSuchProviderException
	 * @throws NoSuchAlgorithmException
	 * @throws CertificateException
	 * @throws IOException
	 */
	private static KeyStore loadKeyStore(String storeFile, char[] password) throws KeyStoreException, NoSuchProviderException, 
	NoSuchAlgorithmException, CertificateException, IOException {
		
		KeyStore keyStore = KeyStore.getInstance("HDImageStore","JCP");
		keyStore.load( (storeFile == null) ? null : new FileInputStream(storeFile), password);
		
		return keyStore;
	}
	
	/**
	 * Signs PDF file (CryptoPro PDF / PKCS7 DETACHED) and saves it.
	 * @param fileToSign - source PDF file.
	 * @param signedFile - signed PDF file.
	 * @param signStore - store containing key to sign file.
	 * @param trustedStore - store containing trusted CA certificate.
	 * @throws IOException
	 * @throws DocumentException
	 * @throws KeyStoreException
	 * @throws NoSuchProviderException
	 * @throws NoSuchAlgorithmException
	 * @throws CertificateException
	 * @throws UnrecoverableKeyException
	 */
	public static void signPDF(String fileToSign, String signedFile, KeyStore signStore, KeyStore trustedStore) throws IOException, 
	DocumentException, KeyStoreException, NoSuchProviderException, NoSuchAlgorithmException, CertificateException, 
	UnrecoverableKeyException {
		
		// Get key and certificates
		PrivateKey key = (PrivateKey)signStore.getKey(ALIAS, PASSWORD);
		
		int chainLength = (trustedStore == null) ? 1 : 2;
		
		Certificate[] chain = new Certificate[chainLength];
		chain[0] = signStore.getCertificate(ALIAS);
		
		if (trustedStore != null) {
			chain[1] = trustedStore.getCertificate(CA_ALIAS);
		}
		
		// Load PDF
		PdfReader reader = new PdfReader(fileToSign);
		FileOutputStream fout = new FileOutputStream(signedFile);
		PdfStamper stp = PdfStamper.createSignature(reader, fout, '\0');
		
		// Sign document
		PdfSignatureAppearance sap = stp.getSignatureAppearance();
		sap.setProvider("JCP");
		sap.setCrypto(key, chain, null, PdfSignatureAppearance.CRYPTOPRO_SIGNED);
		sap.setReason("I'm the signer");
		sap.setLocation("CryptoPro");
		stp.close();

	}
	
	/**
	 * Verifies PKCS7 (CryptoPro PDF / PKCS7 DETACHED) of signed PDF file.
	 * @param fileToVerify - source PDF file to be verified.
	 * @param trustedStore - store containing trusted CA certificate to verify client certificate.
	 * @param crl - CRL file to verify certificate chain.
	 * @return true if signature is valid otherwise throws exception.
	 * @throws IOException
	 * @throws SignatureException
	 * @throws NoSuchAlgorithmException
	 * @throws KeyStoreException
	 * @throws NoSuchProviderException
	 * @throws CertificateException
	 * @throws CRLException
	 */
	public static boolean verifyPDF(String fileToVerify, KeyStore trustedStore, CRL crl) throws IOException, SignatureException, NoSuchAlgorithmException, 
	KeyStoreException, NoSuchProviderException, CertificateException, CRLException {
		
		// CRL collection
		List<CRL> crls = null;
		
		if (crl != null) {
			crls = new ArrayList<CRL>(1);
			crls.add(crl);
		}
		
		boolean result = false;
		
		// Load document
		PdfReader checker = new PdfReader(fileToVerify);
		AcroFields af = checker.getAcroFields();
		
		// Get signatures
		ArrayList<String> names = af.getSignatureNames();
	
		for (int k = 0; k < names.size(); ++k) {
		    
			String name = (String)names.get(k);
			
			System.out.println("Signature: " + name);
			
			// Get one signature and verify it
		    PdfPKCS7 pk = af.verifySignature(name, "JCP");
		    result = pk.verify();
		    
		    // Signer certificate
		    System.out.println("Signer certificate DN: " + pk.getSigningCertificate().getSubjectDN());

		    // Get time stamp and chain
		    Calendar cal = pk.getSignDate();
		    X509Certificate pkc[] = (X509Certificate[])pk.getSignCertificateChain();
		    
		    TimeStampToken ts = pk.getTimeStampToken();
		    if (ts != null) {
		        boolean impr = pk.verifyTimestampImprint();
		        System.out.println("Timestamp imprint verifies: " + impr);
		        cal = pk.getTimeStampDate();
		        System.out.println("Timestamp date: " + cal);
		    }
		    
		    System.out.println("Subject: " + PdfPKCS7.getSubjectFields(pk.getSigningCertificate()));
		    System.out.println("Document modified: " + !result);
		    
		    // Check certificate chain
		    Object fails[] = PdfPKCS7.verifyCertificates(pkc, trustedStore, crls, cal);
		    if (fails == null)
		        System.out.println("Certificates verified against the KeyStore");
		    else
		        System.out.println("Certificate failed: " + fails[1]);
		}
		
		return result;
	}
	
	/**
	 * @param args
	 * @throws NoSuchProviderException 
	 * @throws KeyStoreException 
	 * @throws IOException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws UnrecoverableKeyException 
	 * @throws DocumentException 
	 * @throws SignatureException 
	 * @throws CRLException 
	 */
	public static void main(String[] args) throws KeyStoreException, NoSuchProviderException, NoSuchAlgorithmException, 
	CertificateException, IOException, UnrecoverableKeyException, DocumentException, SignatureException, CRLException {

		// CRL
		CRL crl = null;
		// Trusted store (CA) to build chain
		KeyStore trustedStore = null;
		
		if (CRL_FILENAME != null) {
			crl = CertificateFactory.getInstance("X.509").generateCRL(new FileInputStream(PREFIX_KEY_DIR + CRL_FILENAME));
		}
		
		// Key store containing trusted certificate to build chain
		if (CA_FILENAME != null) {
			trustedStore = loadKeyStore(PREFIX_KEY_DIR + CA_FILENAME, CA_PASSWORD);
		}
		
		// Key store containing client key to sign document
		final KeyStore clientStore = loadKeyStore(null, PASSWORD);
		
		// 1) Sign PDF
		signPDF(PREFIX_PDF_DIR + PDF_FILENAME, 
				PREFIX_PDF_DIR + "signed_" + PDF_FILENAME, clientStore, trustedStore );
		
		// 2) Verify signature
		System.out.println( "Verified: " + verifyPDF(PREFIX_PDF_DIR + "signed_" + PDF_FILENAME, trustedStore, crl ) );

	}
}
